package ru.landgrafhomyak.db.serdha0.api.queries

import ru.landgrafhomyak.db.serdha0.api.misc.BuiltinExpressionsProvider
import ru.landgrafhomyak.db.serdha0.api.misc.Column
import ru.landgrafhomyak.db.serdha0.api.misc.IntermediateColumn
import ru.landgrafhomyak.db.serdha0.api.misc.DatabaseType
import ru.landgrafhomyak.db.serdha0.api.misc.Expression
import ru.landgrafhomyak.db.serdha0.api.misc.FiltersScope
import ru.landgrafhomyak.db.serdha0.api.misc.QueryParam
import ru.landgrafhomyak.db.serdha0.api.misc._Selectable
import ru.landgrafhomyak.db.serdha0.api.runtime.InputRow

public interface SelectQuery<qUE : Any> : _Query<qUE> {
	public enum class Order {
		ASC, DESC
	}

	public class JoinType(
		public val leftNull: LeftNull,
		public val join: Join
	) {
		public enum class LeftNull {
			AS_ID, SKIP, JOIN_TO_NULL
		}

		public enum class Join {
			LEFT_TO_FIRST_RIGHT, LEFT_FOR_EACH_RIGHT, LEFT_FOR_EACH_RIGHT_OR_NULL,
		}
	}

	@Suppress("ClassName")
	public interface _CommonSelectCreatorScope<qUE : Any> {
		public val types: DatabaseType.Provider
		public fun <RT, DT : DatabaseType<RT>> queryParam(name: String, type: DT): QueryParam<RT, DT, qUE>
		public fun <RT, DT : DatabaseType<RT>> paramAsColumn(param: QueryParam<RT, DT, qUE>): IntermediateColumn<RT, DT, qUE>

		public val builtinExpressions: BuiltinExpressionsProvider<qUE>
		public fun <eUE : Any> mapColumns(
			expression: Expression<eUE>,
			input: Expression.InputLinker<eUE, qUE>,
			output: Expression.OutputLinker<eUE, qUE>
		)

		public val filters: FiltersScope<qUE>
		public fun orderBy(order: Order, vararg column: IntermediateColumn<*, *, qUE>)

		public var limit: UInt
		public var offset: ULong

		public fun <RT, DT : DatabaseType<RT>> returnParam(name: String, param: QueryParam<RT, DT, qUE>): Column<RT, DT, qUE>
		public fun <RT, DT : DatabaseType<RT>> returnColumn(name: String, column: IntermediateColumn<RT, DT, qUE>): Column<RT, DT, qUE>
	}

	public interface WithJoin<qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Creator<qUE : Any, lsqUE : Any, rsqUE : Any> : _CommonSelectCreatorScope<qUE> {
			public val left: _Selectable<lsqUE>
			public val right: _Selectable<rsqUE>
			public val joinType: JoinType
			public val referenceColumn: Column<*, *, lsqUE>
			public val targetColumn: Column<*, *, rsqUE>

			public fun createSelectWithJoin(creator: Scope<qUE, lsqUE, rsqUE>): qUE

			public interface Scope<qUE : Any, lsqUE : Any, rsqUE : Any> {
				public val lsqUExt: lsqUE
				public val rsqUExt: rsqUE

				public val leftSubqueryParams: InputRow.WithRedirect<lsqUE, qUE>
				public val rightSubqueryParams: InputRow.WithRedirect<rsqUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromLeft(column: Column<RT, DT, lsqUE>): IntermediateColumn<RT, DT, qUE>
				public fun <RT, DT : DatabaseType<RT>> selectColumnFromRight(column: Column<RT, DT, rsqUE>): IntermediateColumn<RT, DT, qUE>
				public fun <RT : Any, DT : DatabaseType<RT>> selectColumnFromRightAsNullable(column: Column<RT, DT, rsqUE>): IntermediateColumn<RT?, DatabaseType.Nullable<RT, DT>, qUE>
			}
		}
	}

	public interface FromTable<qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Creator<qUE : Any, tUE : Any> {
			public val table: _Selectable<tUE>

			public fun createSelect(creator: Scope<qUE, tUE>): qUE

			public interface Scope<qUE : Any, tUE : Any> : _CommonSelectCreatorScope<qUE> {
				public val tUExt: tUE
				public val subqueryParams: InputRow.WithRedirect<tUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromSubquery(param: Column<RT, DT, tUE>): IntermediateColumn<RT, DT, qUE>
			}
		}
	}
}