package ru.landgrafhomyak.db.serdha0.api.table

import kotlin.jvm.JvmName
import ru.landgrafhomyak.db.serdha0.api.misc.Column
import ru.landgrafhomyak.db.serdha0.api.misc.DatabaseType

/**
 * Descriptor of a reference to row in another table (== foreign key to `INTEGER PRIMARY KEY AUTOINCREMENT`). Used for schema manipulations.
 *
 * @param ctUE Type of [owner table's][ForeignRowReference.fromTable] user expression for static reporting errors when this descriptor passed to wrong table.
 * @param ttUE Type of [target table's][ForeignRowReference.toTable] user expression for static reporting errors when this descriptor passed to wrong table.
 */
public interface ForeignRowReference<ctUE : Any, ttUE : Any, RT, DT : DatabaseType<RT>> : Column<RT, DT, ctUE> {
	/**
	 * Table that contains references. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("fromTable")
	public val fromTable: Table<ctUE, *>

	/**
	 * Table referenced by this foreign key. For debugging.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("toTable")
	public val toTable: Table<ttUE, *>


	/**
	 * Enum of actions to do when referenced row in [ForeignRowReference.toTable] deleted.
	 */
	public enum class OnDelete {
		/**
		 * Forbids deleting row in [ForeignRowReference.toTable] if there is at least one reference from [ForeignRowReference.fromTable].
		 */
		RESTRICT,

		/**
		 * Replaces all references to deleted row in [ForeignRowReference.fromColumn] with `null`.
		 *
		 * Type of [ForeignRowReference.fromColumn] must be [nullable][DatabaseType.Builder.nullableOf].
		 */
		SET_NULL,

		/**
		 * Deletes all rows in [ForeignRowReference.fromTable] that are referencing deleted row in [ForeignRowReference.toTable].
		 */
		CASCADE
	}
}